/*------------------------------------------------------------------------*/
/* Copyright (C) 2015 Armin Biere Johannes Kepler University Linz Austria */
/*------------------------------------------------------------------------*/

#include "common.h"

/*------------------------------------------------------------------------*/

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

/*------------------------------------------------------------------------*/

#include <sys/time.h>
#include <sys/resource.h>

/*------------------------------------------------------------------------*/

int start[2], goal[2], box[4];
int * segs, * top_segs;

/*------------------------------------------------------------------------*/

static int * end_segs;
static int * vsegs, * vtop;
static int * hsegs, * htop;

/*------------------------------------------------------------------------*/

static const char * program_name;
static int common_maxaccel;

/*------------------------------------------------------------------------*/

static size_t allocated_bytes;
size_t max_allocated_bytes;

/*------------------------------------------------------------------------*/
// 'is_valid_move' statistics

static unsigned long checked_moves;
static unsigned long valid_moves;
static unsigned long invalid_moves;

static unsigned long checked_intersects;
static unsigned long checked_intersects_segments;

#ifdef HITTING_SEGMENTS
static unsigned long checked_hitting;
static unsigned long checked_hitting_segments;
static unsigned long hitting_segment;
#endif

static unsigned long out_of_box;
static unsigned long intersected;
static unsigned long hitting_border;

/*------------------------------------------------------------------------*/

static int do_not_print_verbose_message;

static int use_wall_clock_time;

#ifndef NLOG
static int logging_enabled;
void set_logging_enabled () { logging_enabled = 1; }
int is_logging_enabled () { return logging_enabled; }
#endif

/*------------------------------------------------------------------------*/

static void inc_allocated_bytes (size_t bytes) {
  allocated_bytes += bytes;
  assert (bytes <= allocated_bytes);
  if (allocated_bytes > max_allocated_bytes)
    max_allocated_bytes = allocated_bytes;
}

static void dec_alllocated_bytes (size_t bytes) {
  assert (allocated_bytes >= bytes);
  allocated_bytes -= bytes;
}

void * allocate (size_t bytes) {
  void * res = malloc (bytes);
  if (bytes && !res) die ("out-of-memory allocating %lu bytes", bytes);
  inc_allocated_bytes (bytes);
  return res;
}

void * callocate (size_t bytes) {
  void * res = calloc (bytes, 1);
  if (bytes && !res) die ("out-of-memory allocating %lu bytes", bytes);
  inc_allocated_bytes (bytes);
  return res;
}

void * reallocate (void * p, size_t old_bytes, size_t new_bytes) {
  void * res;
  dec_alllocated_bytes (old_bytes);
  res = realloc (p, new_bytes);
  if (new_bytes && !res)
    die ("out-of-memory reallocating %lu bytes", new_bytes);
  inc_allocated_bytes (new_bytes);
  return res;
}

void deallocate (void * p, size_t bytes) {
  dec_alllocated_bytes (bytes);
  free (p);
}

/*------------------------------------------------------------------------*/

void set_program_name (const char * name) { program_name = name; }

void set_common_maxaccel (int m) { common_maxaccel = m; }

void set_quiet_mode () { do_not_print_verbose_message = 1; }

void set_use_wall_clock_time () { use_wall_clock_time = 1; }

/*------------------------------------------------------------------------*/

void msg (const char * fmt, ...) {
  va_list ap;
  if (do_not_print_verbose_message) return;
  if (program_name) printf ("[%s] ", program_name);
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
  fflush (stdout);
}

void die (const char * fmt, ...) {
  va_list ap;
  fputs ("*** ", stdout);
  if (program_name) printf ("%s: ", program_name);
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
  fflush (stdout);
#ifdef VISUALIZE
  void reset_visualization ();
  reset_visualization ();
#endif
  exit (1);
}

/*------------------------------------------------------------------------*/

double average (double a, double b) { return b ? a / b : 0; }
double percent (double a, double b) { return average (100.0 * a, b); }

const char * b2s (size_t bytes) {
  static char buffer[20];
  double b = bytes;
       if (bytes < (1l<<10)) sprintf (buffer, "%lu bytes", bytes);
  else if (bytes < (1l<<20)) sprintf (buffer, "%.1f KB", b/(1l<<10));
  else if (bytes < (1l<<30)) sprintf (buffer, "%.1f MB", b/(1l<<20));
                        else sprintf (buffer, "%.1f GB", b/(1l<<30));
  return buffer;
}

/*------------------------------------------------------------------------*/

static double get_wall_clock_time () {
  struct timeval tv;
  double res;
  if (gettimeofday (&tv, 0)) return 0;
  res = 1e-6 * tv.tv_usec;
  res += tv.tv_sec;
  return res;
}

static double get_process_time () {
  struct rusage u;
  double res;
  if (getrusage (RUSAGE_SELF, &u)) return 0;
  res = u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  return res;
}

double get_time () { 
  if (use_wall_clock_time) return get_wall_clock_time ();
  else return get_process_time ();
}

/*------------------------------------------------------------------------*/

static void check_parcours () {
#ifndef NDEBUG
  int * s;
  assert (0 <= box[0]);
  assert (0 <= box[1]);
  assert (box[0] <= box[2]);
  assert (box[1] <= box[3]);
  assert (box[0] <= start[0]), assert (start[0] <= box[2]);
  assert (box[1] <= start[1]), assert (start[1] <= box[3]);
  assert (box[0] <= goal[0]), assert (goal[0] <= box[2]);
  assert (box[1] <= goal[1]), assert (goal[1] <= box[3]);
  for (s = segs; s < top_segs; s += 2)
    assert (box[0] <= s[0]), assert (s[0] <= box[2]),
    assert (box[1] <= s[1]), assert (s[1] <= box[3]);
  for (s = vsegs; s < vtop; s += 2)
    assert (box[0] <= s[0]), assert (s[0] <= box[2]),
    assert (box[1] <= s[1]), assert (s[1] <= box[3]);
  for (s = hsegs; s < htop; s += 2)
    assert (box[0] <= s[0]), assert (s[0] <= box[2]),
    assert (box[1] <= s[1]), assert (s[1] <= box[3]);
  assert (top_segs - segs == vtop - vsegs + htop - hsegs);
  for (s = segs; s < top_segs; s += 4)
    assert (s[0] <= s[2]), assert (s[1] <= s[3]);
  for (s = vsegs; s < vtop; s += 4)
    assert (s[0] == s[2]), assert (s[1] <= s[3]);
  for (s = hsegs; s < htop; s += 4)
    assert (s[0] <= s[2]), assert (s[1] == s[3]);
  for (s = vsegs + 4; s < vtop; s += 4)
    assert (s[-4] <= s[0]);
  for (s = hsegs + 4; s < htop; s += 4)
    assert (s[-3] <= s[1]);
#endif
}

static int cmp_horizontal_segment_qsort (const void * p, const void * q) {
  const int * a = (const int *) p;
  const int * b = (const int *) q;
  int res;
  assert (a[1] == a[3]);
  assert (b[1] == b[3]);
  res = (a[1] - b[1]);
  if (res) return res;
  return a[0] - b[0];
}

static int cmp_vertical_segment_qsort (const void * p, const void * q) {
  const int * a = (const int *) p;
  const int * b = (const int *) q;
  int res;
  assert (a[0] == a[2]);
  assert (b[0] == b[2]);
  res = (a[0] - b[0]);
  if (res) return res;
  return a[1] - b[1];
}

static void init_horizontal_segments () {
  size_t bytes, count = 0;
  const int * p;
  int * s, * q;
  for (s = segs; s < top_segs; s += 4)
    if (s[1] == s[3] && s[0] != s[2]) count++;
    else assert (s[0] == s[2]);
  if (!count) goto DO_NOT_ALLOCATE;
  bytes = 4 * count * sizeof *hsegs;
  hsegs = allocate (bytes);
  htop = hsegs + 4 * count;
  for (p = segs, q = hsegs; p < top_segs; p += 4)
    if (p[1] == p[3] && p[0] != p[2])
      *q++ = p[0], *q++ = p[1], *q++ = p[2], *q++ = p[3];
  assert (q == htop);
  qsort (hsegs, count, 4 * sizeof *hsegs, cmp_horizontal_segment_qsort);
DO_NOT_ALLOCATE:
  msg ("found and sorted %lu horizontal segments", count);
}

static void init_vertical_segments () {
  size_t bytes, count = 0;
  const int * p;
  int * s, * q;
  for (s = segs; s < top_segs; s += 4)
    if (s[0] == s[2]) count++;
    else assert (s[1] == s[3]);
  if (!count) goto DO_NOT_ALLOCATE;
  bytes = 4 * count * sizeof *vsegs;
  vsegs = allocate (bytes);
  vtop = vsegs + 4 * count;
  for (p = segs, q = vsegs; p < top_segs; p += 4)
    if (p[0] == p[2])
      *q++ = p[0], *q++ = p[1], *q++ = p[2], *q++ = p[3];
  assert (q == vtop);
  qsort (vsegs, count, 4 * sizeof *vsegs, cmp_vertical_segment_qsort);
DO_NOT_ALLOCATE:
  msg ("found and sorted %lu vertical segments", count);
}

static void init_segements_and_check_parcours () {
  init_horizontal_segments ();
  init_vertical_segments ();
  check_parcours ();
}

/*------------------------------------------------------------------------*/

static char parse_error_buffer[256];
static int parsed_numbers, parsed_lines, lineno = 1;

static const char * parse_error (const char * fmt, ...) {
  va_list ap;
  sprintf (parse_error_buffer,
    "line %d after reading %d numbers: ",
    lineno, parsed_numbers);
  va_start (ap, fmt);
  vsprintf (parse_error_buffer + strlen (parse_error_buffer), fmt, ap);
  va_end (ap);
  assert (strlen (parse_error_buffer) + 1 < sizeof parse_error_buffer);
  return parse_error_buffer;
}

#define ERR return parse_error

static int next_char (FILE * file) {
  int ch = getc (file);
  if (ch == '\n') lineno++;
  return ch;
}

#define PARSE(N,S) \
do { \
  int CH; \
  while (isspace (CH = next_char (file))) \
    ; \
  if (CH == EOF) { \
    if (S) ERR (S); \
    goto DONE; \
  } \
  if (!isdigit (CH)) ERR ("expected digit or white space"); \
  N = CH - '0'; \
  while (isdigit (CH = next_char (file))) \
    N = 10*N + (CH - '0'); \
  if (!isspace (CH)) ERR ("expected white space after coordinate"); \
  parsed_numbers++; \
} while (0)

/*------------------------------------------------------------------------*/

const char * read_parcours (FILE * file) {
  for (;;) {
    int x, y, w, h;
    PARSE (x, 0);
    PARSE (y, "one number missing (expected multiple of 4)");
    PARSE (w, "two numbers missing (expected multiple of 4)");
    PARSE (h, "three numbers missing (expected multiple of 4)");
    if (++parsed_lines == 1) {
      if (x > w)
        ERR ("expected box[0] = %d to be at least box[2] = %d", x, w);
      if (y > h)
	ERR ("expected box[1] = %d to be at least box[3] = %d", y, h);
      box[0] = x, box[1] = y, box[2] = w, box[3] = h;
      msg ("bounding-box %d %d %d %d", box[0], box[1], box[2], box[3]);
    } else if (parsed_lines == 2) {
      if (x < box[0] || x > box[2])
	ERR ("start x-coordinate %d not in bounding box", x);
      if (y < box[1] || y > box[3])
	ERR ("start y-coordinate %d not in bounding box", y);
      if (w < box[0] || w > box[2])
	ERR ("goal x-coordinate %d not in bounding box", w);
      if (h < box[1] || h > box[3])
	ERR ("goal y-coordinate %d not in bounding box", h);
      start[0] = x, start[1] = y, goal[0] = w, goal[1] = h;
      msg ("start-goal %d %d %d %d", start[0], start[1], goal[0], goal[1]);
    } else {
      if (x < box[0] || x > box[2])
	ERR ("first x-coordinate %d out of bounding box", x);
      if (y < box[1] || y > box[3])
	ERR ("first y-coordinate %d out of bounding box", y);
      if (w < box[0] || w > box[2])
	ERR ("second x-coordinate %d out of bounding box", w);
      if (h < box[1] || h > box[3])
	ERR ("second y-coordinate %d out of bounding box", h);
      if (x != w && y != h)
	ERR ("expected either x or y coordinates to be the same");
      if (top_segs == end_segs) {
	size_t count = end_segs - segs, size, old_bytes, new_bytes;
	size = count ? 2*count : 4;
	assert (!(size & 3));
	old_bytes = count * sizeof *segs;
	new_bytes = size * sizeof *segs;
	segs = reallocate (segs, old_bytes, new_bytes);
	end_segs = segs + size;
	top_segs = segs + count;
      }
      assert (x == w || y == h);
      if (x > w) { assert (y == h); SWAP (int, x, w); }
      if (y > h) { assert (x == w); SWAP (int, y, h); }
      *top_segs++ = x;
      *top_segs++ = y;
      *top_segs++ = w;
      *top_segs++ = h;
      assert (top_segs <= end_segs);
      // msg ("segment[%lu] %d %d %d %d", (top_segs - segs)/4, x, y, w, h);
    }
  }
DONE:
  if (parsed_lines < 1) ERR ("start and goal coordinates missing");
  if (parsed_lines < 2) ERR ("bounding box missing");
  init_segements_and_check_parcours ();
  return 0;
}

/*------------------------------------------------------------------------*/

void set_default_parcours () {
  int * s;
  box[0] = 0, box[1] = 0, box[2] = 500, box[3] = 400;
  start[0] = 120, start[1] = 180, goal[0] = 320, goal[1] = 220;
  s = segs = calloc (28, sizeof *segs), top_segs = end_segs = s + 28;
  *s++ = 100; *s++ = 100; *s++ = 100; *s++ = 200;
  *s++ = 100; *s++ = 100; *s++ = 200; *s++ = 100;
  *s++ = 100; *s++ = 200; *s++ = 200; *s++ = 200;
  *s++ = 250; *s++ = 200; *s++ = 250; *s++ = 300;
  *s++ = 300; *s++ = 100; *s++ = 300; *s++ = 300;
  *s++ = 300; *s++ = 100; *s++ = 400; *s++ = 100;
  *s++ = 300; *s++ = 200; *s++ = 400; *s++ = 200;
  assert (s == top_segs);
  init_segements_and_check_parcours ();
}

/*------------------------------------------------------------------------*/

void print_check_valid_move_statistics () {
  msg ("checked %lu moves, valid %.0f%%, invalid %.0f%%",
    checked_moves,
    percent (valid_moves, checked_moves),
    percent (invalid_moves, checked_moves));
  msg ("intersection checks %.0f%% with on average %.2f segments",
    percent (checked_intersects, checked_moves),
    average (checked_intersects_segments, checked_intersects));
#ifdef HITTING_SEGMENTS
  msg ("hitting checks %.0f%% with on average %.2f segments",
    percent (checked_hitting, checked_moves),
    average (checked_hitting_segments, checked_hitting));
#endif
  msg ("invalid %lu moves out-of-box %.0f%%",
    out_of_box, percent (out_of_box, invalid_moves));
  msg ("invalid %lu moves intersecting segment %.0f%%",
    intersected, percent (intersected, invalid_moves));
  msg ("invalid %lu moves hitting border %.0f%%",
    hitting_border, percent (hitting_border, invalid_moves));
#ifdef HITTING_SEGMENTS
  msg ("invalid %lu moves hitting segment %.0f%%",
    hitting_segment, percent (hitting_segment, invalid_moves));
#endif
}

/*------------------------------------------------------------------------*/

const int * is_out_of_box (int cx, int cy, int nx, int ny) {
  if (nx > box[0] && nx < box[2] && ny > box[1] && ny < box[3]) return 0;
  return box;
}

/*------------------------------------------------------------------------*/

static int point_on_segment (int a, int l, int r) {
  if (l <= a && a <= r) return 1;
  if (r <= a && a <= l) return 1;
  return 0;
}

static const int *
move_intersects_vertical_segment (const int * m, const int * s) {
  int x, yl, yu, w, h, tmp;
  checked_intersects_segments++;
  LOGMSG (
    "CHECK MOVE %d %d %d %d INTERSECTS VSEG %d %d %d %d",
    m[0], m[1], m[2], m[3], s[0], s[1], s[2], s[3]);
  assert (m[0] <= m[2]);
  assert (s[0] == s[2]);
  assert (m[0] <= s[0]), assert (s[0] <= m[2]);
  w = m[2] - m[0];
  x = s[0];
  if (!w) {
    if (m[0] != x) return 0;
    assert (s[1] <= s[3]);
    yl = m[1], yu = m[3];
    if (yl > yu) SWAP (int, yl, yu);
    assert (yl <= yu);
    if (yu < s[1]) return 0;
    if (yl > s[3]) return 0;
    return s;
  } else {
    if (m[2] < x) return 0;
    if (m[0] > x) return 0;
    yl = s[1], yu = s[3];
    assert (yl <= yu);
    h = m[3] - m[1];
    tmp = w * m[1] + h * (x - m[0]);
    return point_on_segment (tmp, w * yl, w * yu) ? s : 0;
  }
}

static const int * 
move_intersects_horizontal_segment (const int * m, const int * s) {
  int y, xl, xu, w, h, tmp;
  checked_intersects_segments++;
  LOGMSG (
    "CHECK MOVE %d %d %d %d INTERSECTS HSEG %d %d %d %d",
    m[0], m[1], m[2], m[3], s[0], s[1], s[2], s[3]);
  assert (m[3] >= m[1]);
  assert (s[1] == s[3]);
  assert (m[1] <= s[1]), assert (s[1] <= m[3]);
  h = m[3] - m[1];
  y = s[1];
  if (!h) {
    if (m[1] != y) return 0;
    assert (s[0] <= s[2]);
    xl = m[0], xu = m[2];
    if (xl > xu) SWAP (int, xl, xu);
    assert (xl <= xu);
    if (xu < s[0]) return 0;
    if (xl > s[2]) return 0;
    return s;
  } else {
    if (m[3] < y) return 0;
    if (m[1] > y) return 0;
    xl = s[0], xu = s[2];
    assert (xl <= xu);
    w = m[2] - m[0];
    tmp = h * m[0] + w * (y - m[1]);
    return point_on_segment (tmp, h * xl, h * xu) ? s : 0;
  }
}

static const int *
move_intersects_vertical_segments (int cx, int cy, int nx, int ny) {
  const int m[4] = { cx, cy, nx, ny };
  const int * s, * b, * e;
  assert (cx <= nx);
  for (b = vsegs; b != vtop && cx > b[0]; b += 4)
    ;
  for (e = vtop; e != b && nx < e[-4]; e -= 4)
    ;
  for (s = b; s != e; s += 4)
    if (move_intersects_vertical_segment (m, s))
      return s;
  return 0;
}

static const int *
move_intersects_horizontal_segments (int cx, int cy, int nx, int ny) {
  const int m[4] = { cx, cy, nx, ny };
  const int * s, * b, * e;;
  assert (cy <= ny);
  for (b = hsegs; b != htop && cy > b[1]; b+= 4)
    ;
  for (e = htop; e != b && ny < e[-3]; e -= 4)
    ;
  for (s = b; s < e; s += 4)
    if (move_intersects_horizontal_segment (m, s))
      return s;
  return 0;
}

static const int * is_intersected (int cx, int cy, int nx, int ny) {
  const int * res;
  checked_intersects++;
  if (cx < nx) res = move_intersects_vertical_segments (cx, cy, nx, ny);
          else res = move_intersects_vertical_segments (nx, ny, cx, cy);
  if (res) return res;
  if (cy < ny) res = move_intersects_horizontal_segments (cx, cy, nx, ny);
          else res = move_intersects_horizontal_segments (nx, ny, cx, cy);
  return res;
}

/*------------------------------------------------------------------------*/

static int * min_break_dist_table;
static int size_min_break_dist_table;

static int min_break_dist (int v) {
  int res, u;
  assert (v > 0);
  if (v <= common_maxaccel) return 0;
  if (v < size_min_break_dist_table) {
    res = min_break_dist_table[v];
    assert (res >= 0);
    if (res) return res;
  } else res = 0;
  assert (!res);
  for (u = v; u > common_maxaccel; u -= common_maxaccel)
    res += u;
  assert (res > 0);
  LOGMSG ("MINBREAK %d of %d", res, v);
  if (v >= size_min_break_dist_table) {
    size_t old_size = size_min_break_dist_table, new_size = old_size;
    size_t old_bytes, new_bytes;
    if (!new_size) new_size = 1;
    while (new_size <= v) new_size *= 2;
    old_bytes = old_size * sizeof *min_break_dist_table;
    new_bytes = new_size * sizeof *min_break_dist_table;
    min_break_dist_table =
      reallocate (min_break_dist_table, old_bytes, new_bytes);
    while (old_size < new_size)
      min_break_dist_table[old_size++] = 0;
    size_min_break_dist_table = new_size;
  }
  assert (!min_break_dist_table[v]);
  assert (res > 0);
  min_break_dist_table[v] = res;
  return res;
}

/*------------------------------------------------------------------------*/
#ifdef HITTING_SEGMENTS

static const int * 
move_hitting_vseg_posvel (const int * m, const int * s) {
  int vx = m[2] - m[0], vu = m[3] - m[1], vl = vu;
  int  x = m[2],         u = m[3],         l =  u;
  assert (vx > 0);
  assert (s[0] == s[2]);
  assert (s[1] <= s[3]);
  assert (x < s[0]);
  if (x + min_break_dist (vx) < s[0]) return 0;
  checked_hitting_segments++;
  if (u > s[3]) return 0;
  if (l < s[1]) return 0;
  do {
    vx -= common_maxaccel;
    vu += common_maxaccel;
    vl -= common_maxaccel;
    if (vx < 0) return 0;
    x += vx;
    u += vu;
    l += vl;
    if (u > s[3]) return 0;
    if (l < s[1]) return 0;
  } while (x < s[0]);
  return s;
}

static const int *
move_hitting_vsegs_posvel (int cx, int cy, int nx, int ny) {
  const int m[4] = { cx, cy, nx, ny };
  const int * s, * t;
  assert (vsegs);
  assert (cx < nx);
  s = vsegs;
  if (nx >= s[0]) return 0;
  while ((t = s + 4) < vtop && nx < t[0])
    s = t;
  assert (nx < s[0] && (t == vtop || nx >= t[0]));
  return move_hitting_vseg_posvel (m, s);
}

static const int *
move_hitting_vsegs_negvel (int cx, int cy, int nx, int ny) {
  assert (vsegs);
  assert (cx > nx);
  return 0;
}

static const int * 
move_hitting_hseg_posvel (const int * m, const int * s) {
  int vy = m[3] - m[1], vu = m[2] - m[0], vl = vu;
  int  y = m[3],         u = m[2],         l =  u;
  assert (vy > 0);
  assert (s[1] == s[3]);
  assert (s[0] <= s[2]);
  assert (y < s[1]);
  if (y + min_break_dist (vy) < s[1]) return 0;
  checked_hitting_segments++;
  if (u > s[2]) return 0;
  if (l < s[0]) return 0;
  do {
    vy -= common_maxaccel;
    vu += common_maxaccel;
    vl -= common_maxaccel;
    if (vy < 0) return 0;
    y += vy;
    u += vu;
    l += vl;
    if (u > s[2]) return 0;
    if (l < s[0]) return 0;
  } while (y < s[1]);
  return s;
}

static const int *
move_hitting_hsegs_posvel (int cx, int cy, int nx, int ny) {
  const int m[4] = { cx, cy, nx, ny };
  const int * s, * t;
  assert (vsegs);
  assert (cy < ny);
  s = hsegs;
  if (ny >= s[1]) return 0;
  while ((t = s + 4) < htop && ny < t[1])
    s = t;
  assert (ny < s[1] && (t == vtop || nx >= t[1]));
  return move_hitting_hseg_posvel (m, s);
}

static const int *
move_hitting_hsegs_negvel (int cx, int cy, int nx, int ny) {
  assert (hsegs);
  assert (cy > ny);
  return 0;
}

static const int * is_hitting_segment (int cx, int cy, int nx, int ny) {
  const int * res = 0;
  checked_hitting++;
  if (vsegs) {
	 if (cx < nx) res = move_hitting_vsegs_posvel (cx, cy, nx, ny);
    else if (cx > nx) res = move_hitting_vsegs_negvel (cx, cy, nx, ny);
  }
  if (hsegs && !res) {
	 if (cy < ny) res = move_hitting_hsegs_posvel (cx, cy, nx, ny);
    else if (cy > ny) res = move_hitting_hsegs_negvel (cx, cy, nx, ny);
  }
  return res;
}

#endif
/*------------------------------------------------------------------------*/
static const int * is_hitting_border (int cx, int cy, int nx, int ny) {
  int vx = nx - cx, vy = ny - cy;
  if (vx > 0 && nx + min_break_dist (vx) >= box[2]) return box;
  if (vx < 0 && nx - min_break_dist (-vx) <= box[0]) return box;
  if (vy > 0 && ny + min_break_dist (vy) >= box[3]) return box;
  if (vy < 0 && ny - min_break_dist (-vy) <= box[1]) return box;
  return 0;
}

/*------------------------------------------------------------------------*/

#ifndef NLOG
static void
log_violation (const char * name,
               int cx, int cy, int nx, int ny,
	       const int * s) {
  if (logging_enabled)
    msg ("VIOLATION %d %d %d %d %s %d %d %d %d",
      cx, cy, nx, ny, name, s[0], s[1], s[2], s[3]);
}
#else
#define log_violation(...) do { } while (0)
#endif

#define CHECK(NAME) \
do { \
  const int * S = is_ ## NAME (cx, cy, nx, ny); \
  if (!S) break; \
  log_violation (#NAME, cx, cy, nx, ny, S); \
  invalid_moves++; \
  NAME++; \
  return 0; \
} while (0)

/*------------------------------------------------------------------------*/

int is_valid_point (int cx, int cy) {
  const int * s;
  if (cx <= box[0]) return 0;
  if (cx >= box[2]) return 0;
  if (cy <= box[1]) return 0;
  if (cy >= box[3]) return 0;
  for (s = segs; s < top_segs; s += 4)
    if (s[0] == s[2]) {
      if (cx == s[0] && point_on_segment (cy, s[1], s[3])) return 0;
    } else {
      assert (s[1] == s[3]);
      if (cy == s[1] && point_on_segment (cx, s[0], s[2])) return 0;
    }
  return 1;
}

int is_valid_move (int cx, int cy, int nx, int ny) {
  checked_moves++;
  assert (is_valid_point (cx, cy));
  if (cx == nx && cy == ny) goto VALID;
  CHECK (out_of_box);
  CHECK (hitting_border);
  CHECK (intersected);
#ifdef HITTING_SEGMENTS
  CHECK (hitting_segment);
#endif
VALID:
  valid_moves++;
  assert (is_valid_point (nx, ny));
  return 1;
}

/*------------------------------------------------------------------------*/

void reset_parcours () {
  deallocate (min_break_dist_table,
    size_min_break_dist_table * sizeof *min_break_dist_table);
  deallocate (vsegs, (vtop - vsegs) * sizeof *vsegs);
  deallocate (hsegs, (htop - hsegs) * sizeof *hsegs);
  deallocate (segs, (end_segs - segs) * sizeof *segs);
}
