/*------------------------------------------------------------------------*/
/* Copyright (C) 2015 Armin Biere Johannes Kepler University Linz Austria */
/*------------------------------------------------------------------------*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>

/*------------------------------------------------------------------------*/

#include "common.h"
#include "state.h"

/*------------------------------------------------------------------------*/

#ifndef LC				// './configure.sh LC=...'
#define LC 9				// coordinate bits
#endif

#ifndef LV				// './configure.sh LV=...'
#define LV 7				// velocity bits
#endif

#define LR (LV - 5)			// rest velocity bits

#define RC (1<<LC)			// [0..RC[ coordinate range
#define RV (1<<LV)			// [0..RV[ velocity range (unsigned)
#define RR (1<<LR)			// [0..RR[ remaining velocity range

#define MV (RV - 1)			// velocity mask

#define MINV (-(1<<(LV-1)))		// minimum signed velocity
#define MAXV ((1<<(LV-1))-1)		// maximum signed velocity

#define STATES (1ul << (2*LC + 2*LV))	// maximum number of states

/*------------------------------------------------------------------------*/
#if LC > 16
#error "LC larger than 16"     // assume (16 == 8*sizeof (unsigned short))
#endif
/*------------------------------------------------------------------------*/

#define MARKEDBIT(STATE) \
  (assert (((STATE)>>5) < szmarkbits), \
   (markbits[(STATE)>>5] & (1u<<((STATE)&31))))

#define MARKBIT(STATE) \
do { \
  assert (!MARKEDBIT (STATE)); \
  (markbits[(STATE)>>5] |= (1u<<((STATE)&31))); \
  marked++; \
} while (0)

#define MARKBITIFNOTMARKED(STATE) \
do { \
  unsigned * P = markbits + ((STATE)>>5); \
  unsigned BIT = (1u<<((STATE)&31)); \
  assert (((STATE)>>5) < szmarkbits); \
  if ((*P) & BIT) break; \
  *P |= BIT; \
  marked++; \
} while (0)

/*------------------------------------------------------------------------*/

typedef struct Gen Gen;

struct Gen { size_t begin, end; };

/*------------------------------------------------------------------------*/

static const char * name;			// parcour file path

static size_t pushed_states;			// generated states
static size_t max_active_states;		// maximum 'nstates'

static int * accels, * eoa;			// legal acceleration pairs
static size_t accel_bytes;

/*------------------------------------------------------------------------*/
// run-time options

static int incaccel = 1;
static int maxaccel = 10;

static int ignore_moves_that_can_not_be_handled;
static int manhatten_norm, maximum_norm;
static int do_not_sort_generations;
static int garbage_collection_enabled;

#ifdef BACKWARD
static int only_forward, only_backward;
#endif

/*------------------------------------------------------------------------*/
#ifdef VISUALIZE
#include "visualize.h"
static int visualize;
#define LOCK_STATES() \
  do { if (visualize) lock_states (); } while (0)
#define UNLOCK_STATES() \
  do { if (visualize) unlock_states (); } while (0)
#define INIT_VISUALIZATION()  \
  do { if (visualize) init_visualization (), visualize++; } while (0)
#define RESET_VISUALIZATION() \
  do { if (visualize > 1) visualize--, reset_visualization (); } while (0)
#define CHECK_DONE() \
  (visualize ? check_done () : 0)
#else
#define LOCK_STATES() do { } while (0)
#define UNLOCK_STATES() do { } while (0)
#define INIT_VISUALIZATION() do { } while (0)
#define RESET_VISUALIZATION() do { } while (0)
#define CHECK_DONE() (0)
#endif
/*------------------------------------------------------------------------*/
// statistics

static double started;			// start time in seconds
static double expand_time;		// accumulated expanding generations
static double sort_time;		// accumulated sorting generations
static double garbage_collection_time;	// accumulated

static unsigned long generated;		// moves generated
static unsigned long lookups;		// looked up in visited table
static unsigned long hits;		// found already in visited table
static unsigned long ignored;		// moves which can not be handled

static int min_delta_x;			// minimum delta between states
static int min_delta_y;
static int max_delta_x;			// maximum delta between states
static int max_delta_y;

/*------------------------------------------------------------------------*/

static int can_handle_coordinate (int c) { return 0 <= c && c < RC; }
static int can_handle_velocity (int v) { return MINV <= v && v <= MAXV; }

static int can_handle_state (int x, int y, int w, int h) {
  return can_handle_coordinate (x) &&
         can_handle_coordinate (y) &&
         can_handle_velocity (w) &&
	 can_handle_velocity (h);
}

/*------------------------------------------------------------------------*/

static size_t size_generation (Gen * g) { return g->end - g->begin; }

/*------------------------------------------------------------------------*/

// Our radix sort wrt to coordinates is way faster than 'qsort'.
// Preliminary experiments sorting by velocity as well did not
// improve search and further slowed done sorting (since the
// velocity has to be determined during sorting).  We did this
// preliminary test with 'qsort' and for 'radix' sort we would
// actually need a recursive version, e.g., using only RC*RC
// counters is cheap space-wise.

static unsigned counters[RC][RC];

static void sort_generation (const char * type,
                             State * s, Gen * start, Gen * g) {
  size_t n = size_generation (g), bytes, i;
  double entered, end, delta;
  unsigned x, y, pos, count;
  State * tmp;
  entered = get_time ();
  memset (counters, 0, sizeof counters);
  bytes = n * sizeof *tmp;
  tmp = allocate (bytes);
  for (i = g->begin; i < g->end; i++)
    counters[s[i].x][s[i].y]++;
  pos = 0;
  for (x = 0; x < RC; x++)
    for (y = 0; y < RC; y++)
      count = counters[x][y],
      counters[x][y] = pos,
      pos += count;
  for (i = g->begin; i < g->end; i++) {
    State t = s[i];
    tmp[counters[t.x][t.y]++] = t;
  }
  LOCK_STATES ();
  memcpy (s + g->begin, tmp, bytes);
  UNLOCK_STATES ();
  deallocate (tmp, bytes);
  end = get_time ();
  msg ("%.2f: sorted %s %lu of %lu states in %.3f sec",
    end - started, type, g - start, n, (delta = end - entered));
  sort_time += delta;
}

/*------------------------------------------------------------------------*/

static void report_unhandled_state (int x, int y, int w, int h) {
  if (ignore_moves_that_can_not_be_handled) {
    LOGMSG ("IGNORED %d %d %d %d", x, y, w, h);
    ignored++;
  } else if (!can_handle_velocity (w))
    die ("invalid horizontal velocity %d (increase 'LV=%d' or use '-f')",
      w, LV);
  else if (!can_handle_velocity (h))
    die ("invalid vertical velocity %d (increase 'LV=%d' or use '-f')",
      h, LV);
  else if (!can_handle_coordinate (x))
    die ("invalid x-coordinate %d (internal error)", x);
  else
    assert (!can_handle_coordinate (y)),
    die ("invalid y-coordinate %d (internal error)", y);
}

static int print_unknown () {
  printf ("s UNKNOWN\n");
  fflush (stdout);
  return 0;		// SAT Competition exit code UNKNOWN
}

static int print_unreachable () {
  printf ("s UNREACHABLE\n");
  fflush (stdout);
  return 20;		// SAT Competition exit code UNSAT
}

static int cmp_accel_pair_qsort (const void * p, const void * q) {
  const int * a = (const int *) p;
  const int * b = (const int *) q;
  assert (accels <= (int*) a), assert ((int*) a < eoa);
  assert (accels <= (int*) b), assert ((int*) b < eoa);
  if (a[0] < b[0]) return -1;
  if (a[0] > b[0]) return 1;
  if (a[1] < b[1]) return -1;
  if (a[1] > b[1]) return 1;
  return 0;					// unreachable?
}

static void init_accels () {
  size_t max_accel_pairs = (2*maxaccel + 1) * (2*maxaccel + 1), count;
  const long sqrmaxaccel = maxaccel * (long) maxaccel;
  long a, b;
  accel_bytes = max_accel_pairs * 2*sizeof *accels;
  accels = allocate (accel_bytes);
  eoa = accels;
  count = 0;
  for (a = -maxaccel; a <= maxaccel; a += incaccel) {
    for (b = -maxaccel; b <= maxaccel; b += incaccel) {
      if (manhatten_norm) {
	if (abs (a) + abs (b) > maxaccel) continue;
      } else if (maximum_norm) {
	assert (abs (a) <= maxaccel);
	assert (abs (b) <= maxaccel);
      } else {
	long tmp = a*a + b*b;
	if (tmp > sqrmaxaccel) continue;
      }
      LOGMSG ("ADDED ACCEL PAIR %d %d", a, b);
      *eoa++ = a, *eoa++ = b;
      count++;
    }
  }
  assert (count <= 2*max_accel_pairs);
  assert (eoa == accels + 2*count);
  qsort (accels, count, 2*sizeof *accels, cmp_accel_pair_qsort);
#ifndef NLOG
  {
    const int * p;
    for (p = accels; p < eoa; p+= 2)
      LOGMSG ("SORTED ACCEL PAIR %d %d", p[0], p[1]);
  }
#endif
  msg ("prepared acceleration table with %lu pairs %.0f%%",
    count, percent (count, max_accel_pairs));
}

/*------------------------------------------------------------------------*/
#ifdef BACKWARD
static void new_forward_reachable_state (int x, int y, int w, int h);
static void new_backward_reachable_state (int x, int y, int w, int h);
#else
#define new_forward_reachable_state(ARGS...) do { } while (0)
#endif
/*------------------------------------------------------------------------*/

static int forward_target_reached (int, int, int, int);

#define TYPE_GENERATION		"forward"
#define START			start
#define states			forward_states
#define nstates 		forward_nstates
#define szstates 		forward_szstates
#define gens			forward_gens
#define ngens			forward_ngens
#define szgens			forward_szgens
#define visited			forward_visited
#define mark_visited		forward_mark_visited
#define has_been_visited	forward_has_been_visited
#define enlarge_states		forward_enlarge_states
#define statex			forward_statex
#define statey			forward_statey
#define deltax			forward_deltax
#define deltay			forward_deltay
#define push_state		forward_push_state
#define enlarge_generations	forward_enlarge_generations
#define garbage_collection	forward_garbage_collection
#define push_generation		forward_push_generation
#define print_transition	forward_print_transition
#define target_reached		forward_target_reached
#define new_state		forward_new_state
#define new_reachable_state	new_forward_reachable_state
#define last_generation		forward_last_generation
#define end_of_last_generation	forward_end_of_last_generation
#define expand_last_generation	forward_expand_last_generation

#include "search.c"

#undef TYPE_GENERATION
#undef START
#undef states
#undef nstates
#undef szstates
#undef gens
#undef ngens
#undef szgens
#undef visited
#undef mark_visited
#undef has_been_visited
#undef enlarge_states
#undef statex
#undef statey
#undef deltax
#undef deltay
#undef push_state
#undef enlarge_generations
#undef garbage_collection
#undef push_generation
#undef print_transition
#undef target_reached
#undef new_state
#undef new_reachable_state
#undef last_generation
#undef end_of_last_generation
#undef expand_last_generation

/*------------------------------------------------------------------------*/
#ifdef BACKWARD
/*------------------------------------------------------------------------*/

static int backward_target_reached (int, int, int, int);

#define TYPE_GENERATION		"backward"
#define START			goal
#define states			backward_states
#define nstates 		backward_nstates
#define szstates 		backward_szstates
#define gens			backward_gens
#define ngens			backward_ngens
#define szgens			backward_szgens
#define visited			backward_visited
#define mark_visited		backward_mark_visited
#define has_been_visited	backward_has_been_visited
#define enlarge_states		backward_enlarge_states
#define statex			backward_statex
#define statey			backward_statey
#define deltax			backward_deltax
#define deltay			backward_deltay
#define push_state		backward_push_state
#define enlarge_generations	backward_enlarge_generations
#define garbage_collection	backward_garbage_collection
#define push_generation		backward_push_generation
#define print_transition	backward_print_transition
#define target_reached		backward_target_reached
#define new_state		backward_new_state
#define new_reachable_state	new_backward_reachable_state
#define last_generation		backward_last_generation
#define end_of_last_generation	backward_end_of_last_generation
#define expand_last_generation	backward_expand_last_generation

#include "search.c"

#undef TYPE_GENERATION
#undef START
#undef states
#undef nstates
#undef szstates
#undef gens
#undef ngens
#undef szgens
#undef visited
#undef mark_visited
#undef has_been_visited
#undef enlarge_states
#undef statex
#undef statey
#undef deltax
#undef deltay
#undef push_state
#undef enlarge_generations
#undef garbage_collection
#undef push_generation
#undef print_transition
#undef target_reached
#undef new_state
#undef new_reachable_state
#undef last_generation
#undef end_of_last_generation
#undef expand_last_generation

/*------------------------------------------------------------------------*/

#define ONLYVISITED
#define visited			forward_reachable
#define mark_visited		mark_forward_reachable
#define has_been_visited	is_forward_reachable

#include "search.c"

#undef ONLYVISITED
#undef visited
#undef mark_visited
#undef has_been_visited

static void new_forward_reachable_state (int x, int y, int w, int h) {
  if (forward_states && backward_states)
    mark_forward_reachable (x, y, w, h);
}

/*------------------------------------------------------------------------*/

#define ONLYVISITED
#define visited			backward_reachable
#define mark_visited		mark_backward_reachable
#define has_been_visited	is_backward_reachable

#include "search.c"

#undef ONLYVISITED
#undef visited
#undef mark_visited
#undef has_been_visited

static void new_backward_reachable_state (int x, int y, int w, int h) {
  if (forward_states && backward_states)
    mark_backward_reachable (x, y, w, h);
}

/*------------------------------------------------------------------------*/
#endif // BACKWARD
/*------------------------------------------------------------------------*/

static int is_start_state (int x, int y, int w, int h) {
  return x == goal[0] && y == goal[1] && !w && !h;
}

static int forward_target_reached (int x, int y, int w, int h) {
  int res = is_start_state (x, y, w, h);
  if (res) 
    msg ("%.2f: forward target found  %d %d %d %d",
      get_time () - started, x, y, w, h);
#ifdef BACKWARD
  else if (backward_states) {
    int bx = x - w, by = y - h, bw = -w, bh = - h;
    if (is_backward_reachable (bx, by, bw, bh)) {
      msg ("%.2f: forward found matching middle %d %d %d %d",
	get_time () - started, bx, by, bw, bh);
      res = 1;
      hits++;
    }
  }
#endif
  assert (!res || is_valid_point (x, y));
  return res;
}

#ifdef BACKWARD

static int is_goal_state (int x, int y, int w, int h) {
  return x == start[0] && y == start[1] && !w && !h;
}

static int backward_target_reached (int x, int y, int w, int h) {
  int res = is_goal_state (x, y, w, h);
  if (res)
      msg ("%.2f: backward target found %d %d %d %d",
        get_time () - started, x, y, w, h);
  else if (forward_states) {
    int fx = x - w, fy = y - h, fw = -w, fh = - h;
    if (is_forward_reachable (fx, fy, fw, fh)) {
      msg ("%.2f: backward found middle %d %d %d %d",
	get_time () - started, fx, fy, fw, fh);
      res = 1;
      hits++;
    }
  }
  assert (!res || is_valid_point (x, y));
  return res;
}

#endif
/*------------------------------------------------------------------------*/

static int * witness;
static size_t nwitness, szwitness;

static void enlarge_witness () {
  size_t new_szwitness = szwitness ? 2*szwitness : 4;
  size_t old_bytes = szwitness * sizeof *witness;
  size_t new_bytes = new_szwitness * sizeof *witness;
  witness = reallocate (witness, old_bytes, new_bytes);
  szwitness = new_szwitness;
}

static void print_transition (
  int fx, int fy, int fw, int fh,
  int tx, int ty, int tw, int th) {
  const int aw = tw - fw, ah = th - fh;
  printf ("%d %d %d %d --[%d,%d]-> %d %d %d %d (%d,%d)\n",
    fx, fy, fw, fh,
    tw, th,
    tx, ty, tw, th,
    aw, ah);
}

static void push_witness_state (int x, int y, int w, int h) {
  if (nwitness == szwitness) enlarge_witness ();
  witness[nwitness++] = x;
  witness[nwitness++] = y;
  witness[nwitness++] = w;
  witness[nwitness++] = h;
}

static int print_witness () {
  size_t i;
  for (i = 0; i + 4 < nwitness; i += 4) {
    printf ("%lu: ", i/4 + 1);
    print_transition (witness[i+0], witness[i+1], witness[i+2], witness[i+3],
                      witness[i+4], witness[i+5], witness[i+6], witness[i+7]);
  }
  return 10;
}

/*------------------------------------------------------------------------*/
#ifdef BACKWARD

static void push_backward_states (unsigned state) {
  push_witness_state (
    backward_statex (state),
    backward_statey (state),
    backward_deltax (state),
    backward_deltay (state));
  if (state) push_backward_states (backward_states[state].pred);
}

static void fix_backward_witness (size_t s) {
  int i;
  for (i = s; i + 4 < nwitness; i += 4) {
    int w = -witness[i+6];
    int h = -witness[i+7];
    witness[i+4] = witness[i+0] + w;
    witness[i+5] = witness[i+1] + h;
    witness[i+6] = w;
    witness[i+7] = h;
  }
}

static int print_only_backward_witness (unsigned state) {
  printf ("s REACHED\n");
  fflush (stdout);
  if (state) {
    push_backward_states (state);
    fix_backward_witness (0);
  }
  print_witness ();
  fflush (stdout);
  return 10;		// SAT Competition exit code SATISFIABLE
}

static int backward_search () {
  backward_new_state (goal[0], goal[1], 0, 0, 0);
  backward_push_generation ();
  if (backward_target_reached (goal[0], goal[1], 0, 0))
    return print_only_backward_witness (0);
  while (size_generation (backward_last_generation ())) {
    unsigned  state;
    int res = backward_expand_last_generation (&state);
    if (!res) return print_only_backward_witness (state);
    else if (res < 0) return print_unknown ();
  }
  return ignored ? print_unknown () : print_unreachable ();
}

#endif
/*------------------------------------------------------------------------*/

static void push_forward_states (unsigned state) {
  if (state) push_forward_states (forward_states[state].pred);
  push_witness_state (
    forward_statex (state),
    forward_statey (state),
    forward_deltax (state),
    forward_deltay (state));
}

static int print_only_forward_witness (unsigned state) {
  printf ("s REACHED\n");
  fflush (stdout);
  if (state) push_forward_states (state);
  print_witness ();
  fflush (stdout);
  return 10;		// SAT Competition exit code SATISFIABLE
}

static int forward_search () {
  forward_new_state (start[0], start[1], 0, 0, 0);
  forward_push_generation ();
  if (forward_target_reached (start[0], start[1], 0, 0))
    return print_only_forward_witness (0);
  while (size_generation (forward_last_generation ())) {
    unsigned  state;
    int res = forward_expand_last_generation (&state);
    if (!res) return print_only_forward_witness (state);
    else if (res < 0) return print_unknown ();
  }
  return ignored ? print_unknown () : print_unreachable ();
}

/*------------------------------------------------------------------------*/
#ifdef BACKWARD

static int print_forward_meets_backward_witness (unsigned middle) {
  unsigned idx, pred;
  int x, y, w, h;
  size_t s;
  printf ("s REACHED\n");
  fflush (stdout);
  push_forward_states (middle);
  w = -forward_deltax (middle);
  h = -forward_deltay (middle);
  x = forward_statex (middle) + w;
  y = forward_statey (middle) + h;
  for (idx = backward_nstates-1;;idx--) {
    if (backward_statex (idx) == x &&
        backward_statey (idx) == y &&
        backward_deltax (idx) == w &&
        backward_deltay (idx) == h) break;
    assert (idx);
  }
#ifdef VISUALIZE
  LOCK_STATES ();
  backward_nstates = idx + 1;
  if (backward_last_visualized >= backward_nstates)
    backward_last_visualized = idx;
  UNLOCK_STATES ();
#endif
  s = nwitness;
  assert (s > 3);
  pred = backward_states[idx].pred;
  push_backward_states (pred);
  fix_backward_witness (s - 4);
  print_witness ();
  return 10;
}

static int print_backward_meets_forward_witness (unsigned middle) {
  unsigned idx, pred;
  int x, y, w, h;
  size_t s;
  printf ("s REACHED\n");
  fflush (stdout);
  w = -backward_deltax (middle);
  h = -backward_deltay (middle);
  x = backward_statex (middle) + w;
  y = backward_statey (middle) + h;
  for (idx = forward_nstates-1;;idx--) {
    if (forward_statex (idx) == x &&
        forward_statey (idx) == y &&
        forward_deltax (idx) == w &&
        forward_deltay (idx) == h) break;
    assert (idx);
  }
  pred = forward_states[idx].pred;
#ifdef VISUALIZE
  LOCK_STATES ();
  forward_nstates = idx + 1;
  if (forward_last_visualized >= forward_nstates)
    forward_last_visualized = idx;
  UNLOCK_STATES ();
#endif
  push_forward_states (pred);
  s = nwitness;
  assert (s > 3);
  push_backward_states (middle);
  fix_backward_witness (s - 4);
  print_witness ();
  return 10;
}

static int forward_backward_search () {
  forward_new_state (start[0], start[1], 0, 0, 0);
  forward_push_generation ();
  if (forward_target_reached (start[0], start[1], 0, 0))
    return print_only_forward_witness (0);
  backward_new_state (goal[0], goal[1], 0, 0, 0);
  backward_push_generation ();
  if (backward_target_reached (goal[0], goal[1], 0, 0))
    return print_only_backward_witness (0);
  for (;;) {
    unsigned  state;
    int res;
    if (!size_generation (forward_last_generation ())) break;
    res = forward_expand_last_generation (&state);
    if (!res) return print_forward_meets_backward_witness (state);
    else if (res < 0) return print_unknown ();
    if (!size_generation (backward_last_generation ())) break;
    res = backward_expand_last_generation (&state);
    if (!res) return print_backward_meets_forward_witness (state);
    else if (res < 0) return print_unknown ();
  }
  return ignored ? print_unknown () : print_unreachable ();
}

#endif
/*------------------------------------------------------------------------*/

static int search () {
  if (!is_valid_point (start[0], start[1])) {
    msg ("start state %d %d invalid", start[0], start[1]);
    return print_unreachable ();
  }
  if (!is_valid_point (goal[0], goal[1])) {
    msg ("goal state %d %d invalid", goal[0], goal[1]);
    return print_unreachable ();
  }
  if (!can_handle_state (start[0], start[1], 0, 0))
    die ("can not handle start state %d %d", start[0], start[1]);
  if (!can_handle_state (goal[0], goal[1], 0, 0))
    die ("can not handle goal state %d %d", goal[0], goal[1]);
#ifdef BACKWARD
  if (only_backward) return backward_search ();
  else if (!only_forward) return forward_backward_search ();
#endif
     return forward_search ();
}

/*------------------------------------------------------------------------*/

static void init_search () {
#ifndef STATICVISITED
#ifdef BACKWARD
  if (!only_backward) 
#endif
    forward_visited = callocate (STATES>>3);
#ifdef BACKWARD
  if (!only_forward) {
    backward_visited = callocate (STATES>>3);
    if (!only_backward) {
      forward_reachable = callocate (STATES>>3);
      backward_reachable = callocate (STATES>>3);
    }
  }
#endif
#endif
  init_accels ();
  msg ("%.2f: search initialized", get_time () - started);
  INIT_VISUALIZATION ();
}

static void init_statistics () {
#ifdef VISUALIZE
  if (visualize) {
    msg ("measuring wall clock time");
    set_use_wall_clock_time ();
  } else
#endif
  msg ("measuring process time");
  started = get_time ();
}

static const char * usage =
"usage: mcracetrack [ <option> ... ] [ <parcours> ]\n"
"\n"
"where <option> is one of the following\n"
"\n"
"-h               print this command line option summary\n"
"-q               quiet mode: do not print verbose messages\n"
#ifndef NLOG     
"-l               enable low-level logging on <stdout>\n"
#endif          
"-a <N>           maximum accelaration (default %d)\n"
"-i <N>           accelaration increment (default %d)\n"
"-f               ignore state which can not be handled\n"
"\n"
"-m | --manhatten | --manhatten-norm\n"
"-M | --maximum   | --maximum-norm\n"
"\n"
#ifdef BACKWARD
"--only-forward   only use forward search\n"
"--only-backard   only use backward search\n"
"                 (default is to use both forward and backward)\n"
"\n"
#endif
"--no-sort        do not sort generations (worse cache locality)\n"
"--do-not-sort\n"
"\n"
#ifdef VISUALIZE
"-v       visualize search\n"
#endif
"\n"
"If no <parcours> file is specified the default parcours of the\n"
"c't-Racetrack as published in c't 2015 issue 23 is used.  Otherwise\n"
"<parcours> should contain first the bottom left and top right\n"
"coordinates of the bounding box, then start and goal coordinates\n"
"followed by line segments. All are non-negative integer quadruples\n"
"with increasing x and y coordinates (except for start and goal).\n"
;

static void print_usage () {
  printf (usage, maxaccel, incaccel);
  fflush (stdout);
}

static void parse_options (int argc, char ** argv) {
  int i;
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-h")) print_usage (), exit (1);
    else if (!strcmp (argv[i], "-a")) {
      if (i + 1 == argc) die ("option to '-a' missing");
      if ((maxaccel = atoi (argv[++i])) <= 0)
	die ("invalid '-a' argument '%s'", argv[i]);
    } else if (!strcmp (argv[i], "-i")) {
      if (i + 1 == argc) die ("option to '-i' missing");
      if ((incaccel = atoi (argv[++i])) <= 0)
	die ("invalid '-i' argument '%s'", argv[i]);
    } 
    else if (!strcmp (argv[i], "-q")) set_quiet_mode ();
#ifndef NLOG
    else if (!strcmp (argv[i], "-l")) set_logging_enabled ();
#endif
    else if (!strcmp (argv[i], "-f"))
      ignore_moves_that_can_not_be_handled = 1;
    else if (!strcmp (argv[i], "-M") ||
             !strcmp (argv[i], "--maximum") ||
             !strcmp (argv[i], "--maximum-norm"))
      maximum_norm = 1;
    else if (!strcmp (argv[i], "-m") ||
             !strcmp (argv[i], "--manhatten") ||
             !strcmp (argv[i], "--manhatten-norm"))
      manhatten_norm = 1;
#ifdef BACKWARD
    else if (!strcmp (argv[i], "--only-forward")) only_forward = 1;
    else if (!strcmp (argv[i], "--only-backward")) only_backward = 1;
#endif
    else if (!strcmp (argv[i], "--no-sort") || 
             !strcmp (argv[i], "--do-not-sort"))
      do_not_sort_generations = 1;
    else if (!strcmp (argv[i], "-g"))
      garbage_collection_enabled = 1;
#ifdef VISUALIZE
    else if (!strcmp (argv[i], "-v"))
      visualize = 1;
#endif
    else if (argv[i][0] == '-')
      die ("invalid option '%s' (try '-h')", argv[i]);
    else if (name)
      die ("multiple parcours '%s' and '%s' (try '-h')", name, argv[i]);
    else name = argv[i];
  }
  if (maximum_norm && manhatten_norm)
    die ("can not use both maximum and manhatten norm");

#ifdef BACKWARD
  if (only_backward && only_forward)
    die ("can not combine '--only-forward' and '--only-backward'");
#endif
}

static void banner () {
  msg ("Explicit McRaceTrack Model Checker Version %s", version_string ());
  msg ("Copyright (C) 2015 Armin Biere JKU Linz");
  msg ("%s", compile_time_options ());
#ifndef NLOG
  if (is_logging_enabled ()) msg ("logging enabled");
  else msg ("logging disabled");
#endif
#ifndef NDEBUG
  msg ("assertion checking enabled");
#endif
  assert (0 < maxaccel);
  assert (0 < incaccel);
  assert (incaccel <= maxaccel);
  if (incaccel > maxaccel) {
    int orig = incaccel;
    incaccel = maxaccel;
    msg ("maximum acceleration %d reduced increment %d (from %d)",
      maxaccel, incaccel, orig);
  } else if ((maxaccel % incaccel)) {
    int orig = maxaccel;
    assert (incaccel < maxaccel);
    do maxaccel--; while (maxaccel % incaccel);
    msg ("maximum acceleration reduced to %d (from %d) increment %d",
      maxaccel, orig, incaccel);
  } else
    msg ("maximum acceleration %d increment %d", maxaccel, incaccel);
  msg ("using %s norm to limit acceleration",
    (manhatten_norm ?
      "manhatten" : (maximum_norm ? "maximum" :"euclidean")));
  set_common_maxaccel (maxaccel);
  msg ("moves which can not be handled will %s",
    ignore_moves_that_can_not_be_handled ? "be ignored" : "force abort");
  msg ("garbage collection %s",
    garbage_collection_enabled ? "enabled" : "disabled");
#ifdef BACKWARD
  if (only_forward)
    msg ("only forward reachability analysis");
  else if (only_backward)
    msg ("only backward reachability analysis");
  else
    msg ("both forward and backward reachability analysis");
#endif
}

static void init_parcours () {
  const char * err;
  FILE * file;
  if (name) {
    if (!(file = fopen (name, "r"))) die ("can not read '%s'", name);
    msg ("reading '%s'", name);
    err = read_parcours (file);
    if (err) die ("parse error in '%s' %s", name, err);
    fclose (file);
  } else {
    msg ("using default parcours of c't 2015 issue 23");
    set_default_parcours ();
  }
}

static void print_sizes () {
  msg ("maximum number of states %lu", STATES);
  msg ("velocity range [0..%d] %d bits", RV - 1, LV);
  msg ("coordinate range [0..%d] %d bits", RC - 1, LC);
  msg ("sizeof (long) = %lu bytes", sizeof (long));
  msg ("sizeof (State) = %lu bytes", sizeof (State));
  msg ("visited table needs %s", b2s (STATES>>3));
  msg ("maximum states stack of %s",
    b2s ((1ul<<32) * sizeof (State)));
  assert (sizeof (size_t) == sizeof (long));
  assert (sizeof (long) == 8);
}

static void print_memory_statistics () {
  size_t a = max_allocated_bytes, t = a, s = sizeof counters;
#ifdef STATICVISITED
  s += sizeof forward_visited;
#ifdef BACKWARD
  s += sizeof backward_visited;
#endif
#endif
  t += s;
  msg ("fixed static memory of %s %.0f%%", b2s (s), percent (s, t));
  msg ("maximum allocated heap memory of %s %.0f%%", b2s (a), percent (a, t));
  msg ("total maximum allocated memory of %s", b2s (t));
}

static void print_search_statistics (double t) {
  size_t ngens = forward_ngens;
  size_t nstates = forward_nstates;
#ifdef BACKWARD
  nstates += backward_nstates;
  ngens += backward_ngens;
#endif
  msg ("worked on %lu generations", ngens);
  msg ("found %lu states %.0f%% at %.3f millions/sec",
    pushed_states,
    percent (pushed_states, STATES),
    average (pushed_states/1e6, t));
  if (garbage_collection_enabled) {
    size_t collected = (pushed_states - nstates);
    msg ("active %lu states %.0f%%, %lu max %.0f%%",
      nstates, percent (nstates, pushed_states),
      max_active_states, percent (max_active_states, pushed_states));
    msg ("collected %s of %lu states %.0f%%",
      b2s (collected * sizeof (State)),
      collected, percent (collected, pushed_states));
  }
  msg ("generated %lu moves, %.3f million moves/sec",
    generated, average (generated/1e6, t));
  msg ("looked up %lu moves %.0f%, %lu hits %.0f%",
    lookups, percent (lookups, generated),
    hits, percent (hits, lookups));
  msg ("moves range from [%d,%d] to [%d,%d]",
    min_delta_x, min_delta_y, max_delta_x, max_delta_y);
  if (ignore_moves_that_can_not_be_handled)
    msg ("ignored %lu moves which can not be handled", ignored);
}

static void print_statistics () {
  double t = get_time () - started;
  print_search_statistics (t);
  print_check_valid_move_statistics (t);
  print_memory_statistics ();
  if (!do_not_sort_generations)
    msg ("sorting time %.2f sec %.0f%%",
      sort_time, percent (sort_time, t));
  if (garbage_collection_enabled)
    msg ("garbage collection time %.2f sec %.0f%%",
      garbage_collection_time, percent (garbage_collection_time, t));
  msg ("expansion time %.2f sec %.0f%%",
    expand_time, percent (expand_time, t));
  msg ("finished after %.2f seconds", t);
}

static void reset_all () {
  RESET_VISUALIZATION ();
#if !defined (STATICVISITED)
#ifdef BACKWARD
  if (!only_backward) 
#endif
    deallocate (forward_visited, STATES>>3);
#ifdef BACKWARD
  if (!only_forward) {
    deallocate (backward_visited, STATES>>3);
    if (!only_backward) {
      deallocate (forward_reachable, STATES>>3);
      deallocate (backward_reachable, STATES>>3);
    }
  }
#endif
#endif
  reset_parcours ();
  deallocate (forward_states, forward_szstates * sizeof*forward_states);
  deallocate (forward_gens, forward_szgens * sizeof *forward_gens);
  deallocate (witness, szwitness * sizeof *witness);
  deallocate (accels, accel_bytes);
}

/*------------------------------------------------------------------------*/

static int catched_signal;
static void (*sig_int_handler)(int);
static void (*sig_segv_handler)(int);
static void (*sig_abrt_handler)(int);
static void (*sig_term_handler)(int);

static void reset_signal_handlers (void) {
  (void) signal (SIGINT, sig_int_handler);
  (void) signal (SIGSEGV, sig_segv_handler);
  (void) signal (SIGABRT, sig_abrt_handler);
  (void) signal (SIGTERM, sig_term_handler);
}

static void catchsig (int sig) {
  if (!catched_signal) {
    catched_signal = 1;
    printf ("*** mcracetrack: caught signal %d\n", sig);
    fflush (stdout);
    print_statistics ();
    printf ("*** mcracetrack: raising caught signal %d\n", sig);
    fflush (stdout);
  }
  reset_signal_handlers ();
  raise (sig);
}

static void init_signal_handlers (void) {
  sig_int_handler = signal (SIGINT, catchsig);
  sig_segv_handler = signal (SIGSEGV, catchsig);
  sig_abrt_handler = signal (SIGABRT, catchsig);
  sig_term_handler = signal (SIGTERM, catchsig);
}

/*------------------------------------------------------------------------*/

int main (int argc, char ** argv) {
  int res;
  set_program_name ("mcracetrack");
  parse_options (argc, argv);
  banner ();
  init_statistics ();
  init_signal_handlers ();
  init_parcours ();
  print_sizes ();
  init_search ();
  res = search ();
  reset_signal_handlers ();
  print_statistics ();
  reset_all ();
  return res;
}
